{
 "cells": [
  {
   "cell_type": "markdown",
   "id": "ecea5660-2d2c-45a2-9ee1-b8de319d55ad",
   "metadata": {},
   "source": [
    "# 2.7 Αποσφαλμάτωση\n",
    "© Γιάννης Κωστάρας\n",
    "\n",
    "---\n",
    "\n",
    "[<](../2.6-IDEs/2.6-IDEs.ipynb) | [Δ](../../TOC.ipynb)\n",
    "\n",
    "---\n",
    "\n",
    "### Μαθησιακοί στόχοι\n",
    "Σε αυτήν την ενότητα θα μάθουμε:\n",
    "\n",
    "* πώς να αποσφαλματώνουμε (debug) τα προγράμματά μας\n",
    "\n",
    "## Εισαγωγή\n",
    "Η λέξη “Bug” (ή _σφάλμα_ στην ελληνική του απόδοση) προέρχεται από τη ναύαρχο Grace Hopper το 1947 καθώς εργαζόταν σε έναν Η/Υ τύπου Mark II στο Harvard. Η δυσλειτουργία του προγράμματος οφειλόταν σ’ ένα ζωύφιο (bug) που είχε κολλήσει σ’ ένα κύκλωμα, οπότε η διαδικασία αποκόλλησής του ονομάστηκε _debugging_. (Wikipedia)\n",
    "\n",
    "Υπάρχουν πολλών ειδών bugs που θα αντιμετωπίσετε στη Java:\n",
    "\n",
    "* **Λάθη μεταγλώττισης (compile-time errors)**: λάθη που τα βρίσκει ο μεταγλωττιστής, π.χ. συντακτικά λάθη (λείπει ένα ; ή μια παρένθεση ()) ) ή λάθη τύπων δεδομένων π.χ. μη δήλωση τύπου μιας μεταβλητής, εκχώριση τιμής σε λάθος τύπο δεδομένων κλπ.  \n",
    "* **Λάθη εκτέλεσης (run-time errors)**: λάθη που τα βρίσκετε κατά την εκτέλεση του προγράμματος, π.χ. διαίρεση με το μηδέν, άνοιγμα ενός αρχείου που δεν υπάρχει κλπ.\n",
    "* **Λογικά λάθη (logic errors)**: λάθη που τα βρίσκει είτε ο προγραμματιστής είτε ο ελεγκτής (tester) του προγράμματος τα οποία είναι είτε λάθος αποτελέσματα είτε λαθεμένη συμπεριφορά του προγράμματος και είναι τα πιο δύσκολα να επιλυθούν. Λέγεται, ότι ένας προγραμματιστής ξοδεύει το 75% του χρόνου του στην επίλυση τέτοιων λαθών.\n",
    "\n",
    "Για την αντιμετώπιση των λογικών λαθών (αλλά και των λαθών εκτέλεσης), μπορείτε να χρησιμοποιήσετε την τεχνική της αποσφαλμάτωσης ή εκσφαλμάτωσης (debugging). «Αποσφαλμάτωση» ή «εκσφαλμάτωση» (debugging) ονομάζεται η μεθοδική διαδικασία εύρεσης και εξάλειψης σφαλμάτων κώδικα ενός προγράμματος υπολογιστή ώστε να συμπεριφέρεται όπως προβλέπεται.\n",
    "\n",
    "Διαδικασία:\n",
    "\n",
    "1. Πίστεψε ότι το πρόβλημα μπορεί να λυθεί και επέμενε μέχρι τη λύση του\n",
    "1. Επανάληψη/επιβεβαίωση του προβλήματος (αν το πρόβλημα είναι αναπαραγώσιμο)\n",
    "1. Απομόνωση του σημείου που εμφανίζεται το σφάλμα (συνήθως από τα αρχεία καταγραφής – logs)\n",
    "1. Αναγνώριση της αιτίας που το προκαλεί\n",
    "1. Διόρθωση του σφάλματος\n",
    "1. Επιβεβαίωση της διόρθωσης\n",
    "\n",
    "Με τη βοήθεια ενός _αποσφαλματωτή (debugger)_, ωθούμε το πρόγραμμα ώστε να εισέλθει σε μια συγκεκριμένη κατάσταση ώστε να εμφανιστεί το πρόβλημα.\n",
    "\n",
    "Το NetBeans περιλαμβάνει έναν visual debugger που διευκολύνει τη διαδικασία.\n",
    "\n",
    "## Αποσφαλμάτωση με το NetBeans\n",
    "Δημιουργήστε ένα νέο έργο (project) με όνομα ```Debug``` στο NetBeans όπως μάθαμε στο πρώτο μάθημα αυτής της εβδομάδας. \n",
    "Επικολλήστε τον παρακάτω κώδικα μέσα στη μέθοδο ```main()```:\n",
    "\n",
    "```java\n",
    "public static void main(String[] args) {\n",
    "   int[] ints = {1, 2, 3, 4, 5};\n",
    "   for (int i = 0; i <= ints.length; i++) {\n",
    "       System.out.println(ints[i]*ints[i]);\n",
    "   }\n",
    "}\n",
    "```\n",
    "\n",
    "Εκτελώντας το πρόγραμμα βλέπουμε το παρακάτω αποτέλεσμα (μενού ```Window --> Output```):\n",
    "```java\n",
    "1\n",
    "4\n",
    "9\n",
    "16\n",
    "25\n",
    "Exception in thread \"main\" java.lang.ArrayIndexOutOfBoundsException: 5\n",
    "```\n",
    "Αν και το πρόγραμμα υπολογίζει σωστά τα τετράγωνα των 5 αριθμών (1-5), στο τέλος εγείρει μια εξαίρεση εκτέλεσης που μας δηλώνει ότι κάπου έχουμε κάνει κάποιο λογικό λάθος.\n",
    "\n",
    "Μπορούμε να ελέγξουμε γραμμή-γραμμή που μπορεί να έχουμε κάνει λάθος, αλλά μπορούμε να χρησιμοποιήσουμε τον οπτικό μεταγλωττιστή του NetBeans.\n",
    "\n",
    "Κάντε κλικ πάνω στον αριθμό γραμμής της πρώτης γραμμής της μεθόδου ```main()``` για να εισάγετε ένα _σημείο παύσης ή σημείο αναστολής (breakpoint)_. Στη συνέχεια πατήστε το κουμπί **Debug** όπως φαίνεται στην παρακάτω εικόνα.\n",
    "\n",
    "![](assets/Fig1.png)\n",
    "\n",
    "**Εικόνα 2.7.1** _Εκκίνηση του αποσφαλματωτή_\n",
    "\n",
    "Η εκτέλεση του προγράμματος θα ξεκινήσει και θα σταματήσει στη γραμμή που έχετε ορίσει το σημείο παύσης. Από δω και πέρα μπορείτε να χρησιμοποιήσετε τα κουμπιά αποσφαλμάτωσης για να μετακινηθείτε γραμμή-γραμμή στον κώδικα, να εξετάσετε τις τιμές των μεταβλητών κλπ. \n",
    "\n",
    "![](assets/Fig2.png)\n",
    "\n",
    "**Εικόνα 2.7.2** _Παύση του αποσφαλματωτή σ' ένα σημείο παύσης_\n",
    "\n",
    "Τα κουμπιά που φαίνονται στην εικόνα 2 είναι τα εξής:\n",
    "\n",
    "* Τερματισμός του αποσφαλματωτή (```Shift+F5```)\n",
    "* Παύση εκτέλεσης\n",
    "* Συνέχεια εκτέλεσης (```F5```)\n",
    "* Εκτέλεση ενός βήματος χωρίς είσοδο μέσα στις μεθόδους, δηλ. οι μέθοδοι εκτελούνται σαν εντολές (```F8```)\n",
    "* Εκτέλεση ενός βήματος χωρίς είσοδο μέσα στις εκφράσεις (```Shift+F8```)\n",
    "* Εκτέλεση ενός βήματος μπαίνοντας μέσα στις μεθόδους (```F7```)\n",
    "* Έξοδος από τη μέθοδο\n",
    "* Εκτέλεση μέχρι εκεί που βρίσκεται ο δρομέας (```F4```)\n",
    "* Εφαρμόζει τις αλλαγές στον κώδικα κατά τη διάρκεια της αποσφαλμάτωσης \n",
    "* Λαμβάνει ένα στιγμιότυπο  "
   ]
  },
  {
   "cell_type": "markdown",
   "id": "704b64c7-4c3f-417b-ad73-e4c90f00b517",
   "metadata": {},
   "source": [
    "## Αποσφαλμάτωση με το Visual Studio Code\n",
    "Αντίστοιχα δουλεύει και η αποσφαλμάτωση στο VSCode. Ορίστε τα σημεία παύσης (breakpoints) όπως και στο NetBeans. Από το μενού **Run** μπορείτε να επιλέξετε **Start Debugging** για αποσφαλμάτωση (αντί για **Run Without Debugging** για εκτέλεση του κώδικα).\n",
    "\n",
    "## Αποσφαλμάτωση με το BlueJ\n",
    "Ορίστε και εδώ σημεία αναστολής (breakpoints) στο κενό στα αριστερά του κειμενογράφου. Όταν τρέξετε την ```main()``` η εκτέλεση θα σταματήσει στο πρώτο σημείο αναστολής και θα εμφανιστεί το παρακάτω παράθυρο με τις μεταβλητές και τις τιμές τους εκείνη τη στιγμή.\n",
    "\n",
    "![](assets/Fig3.png)\n",
    "\n",
    "**Εικόνα 2.7.3** _Αποσφαλματωτής του BlueJ_"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "537f96c2-0089-4a64-96bc-d5b7a8abacf9",
   "metadata": {},
   "source": [
    "## Αποσφαλμάτωση χωρίς ΟΠΕ\n",
    "Αν επιλέξατε να μην εγκαταστήσετε ένα ΟΠΕ στον Η/Υ σας, τότε μπορείτε να χρησιμοποιήσετε αυτόν τον [online αποσφαλματωτή](http://www.pythontutor.com/visualize.html#mode=edit) με τον οποίο μπορείτε επίσης να οπτικοποιήσετε την εκτέλεση του προγράμματός σας βήμα-βήμα.\n",
    "\n",
    "Καλή επιτυχία και καλό κουράγιο κατά την αποσφαλμάτωση των προγραμμάτων σας."
   ]
  },
  {
   "cell_type": "markdown",
   "id": "c3e6177a-4b6b-4d60-a1f1-c6bbc245a27e",
   "metadata": {},
   "source": [
    "## Άσκηση\n",
    "1. Βρείτε τα λάθη στο παρακάτω πρόγραμμα, αναφέρετε τι κατηγορία λάθους είναι το καθένα, και διορθώστε τα ώστε να εκτελεστεί σωστά το πρόγραμμα.\n",
    "\n",
    "```java\n",
    "import java.util.Random;\n",
    "\n",
    "public class FortuneTeller {\n",
    "    \n",
    "    public static void Main(String[] args) {\n",
    "        Random random = new Random();\n",
    "        int fortune = random.nextInt(10);  // τυχαίος ακέραιος μεταξύ 0 και 9\n",
    "        \n",
    "        if (fortune = 0) {\n",
    "            System.out.println(\"Your aims are high, and you are capable of much.\");\n",
    "            \n",
    "        } else if (fortune == 1) {\n",
    "            System.out.println(\"Don't behave with cold manners.\")\n",
    "            \n",
    "        } else if (fortune == 2) {\n",
    "            System.out.println(\"May you someday be carbon neutral\");\n",
    "            \n",
    "        } else if (fortune == 3) {\n",
    "            System.out.println(\"You are special in a way you will soon begin to understand.\");\n",
    "            \n",
    "        } else if (fortune == 4) {\n",
    "            System.out.println(\"A conclusion is simply the place where you got tired of thinking.\");\n",
    "            \n",
    "        } else if (fortune ==) {\n",
    "            System.out.println(\"No snowflake feels responsible in an avalanche.\");\n",
    "            \n",
    "        } else if (fortune == 6) {\n",
    "            System.out.println(\"He who laughs last is laughing at you.\");\n",
    "            \n",
    "        } else if (fortune == 7) {\n",
    "            System.out.println(\"If you look back, you'll soon be going that way.\");\n",
    "            \n",
    "        } else if (fortune == 8) {\n",
    "            System.out.println(\"Face facts with dignity.\");\n",
    "            \n",
    "        } else if (fortune == 9   {\n",
    "            System.out.println(\"The fortune you seek is in another cookie.\");\n",
    "            \n",
    "        }\n",
    "    }\n",
    "}\n",
    "```\n",
    "\n",
    "## Πηγές\n",
    "1. [Αποσφαλμάτωση](https://el.wikipedia.org/wiki/%CE%91%CF%80%CE%BF%CF%83%CF%86%CE%B1%CE%BB%CE%BC%CE%AC%CF%84%CF%89%CF%83%CE%B7), Wikipedia\n",
    "1. Horstmann C., [Big Java 5 - Chapter 6 - Loops](https://www.cs.ryerson.ca/~aferworn/courses/CPS109/CLASSES/week06/ch06/index.html)\n",
    "1. Kolling M. (μετάφραση Ζερβός Τ.), [Το εγχειρίδιο του BlueJ](https://bluej.org/tutorial/tutorial-greek.pdf), έκδοση 1.4, University of Southern Denmark.\n",
    "1. Spinellis D. (2017), _Effective Debugging_, Addison-Wesley.\n",
    "1. [Using the Visual Debugger in NetBeans IDE](https://netbeans.org/kb/docs/java/debug-visual.html)"
   ]
  },
  {
   "cell_type": "markdown",
   "id": "46f7a4aa-e0e8-47ff-a1c1-127cae975428",
   "metadata": {},
   "source": [
    "---\n",
    "\n",
    "[<](../2.6-IDEs/2.6-IDEs.ipynb) | [Δ](../../TOC.ipynb)\n",
    "\n",
    "---"
   ]
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Java",
   "language": "java",
   "name": "java"
  },
  "language_info": {
   "codemirror_mode": "java",
   "file_extension": ".jshell",
   "mimetype": "text/x-java-source",
   "name": "Java",
   "pygments_lexer": "java",
   "version": "17+35-LTS-2724"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 5
}
